@@ -1,342 +1,42 @@
 package Heap::Simple;
-use strict;
 use Carp;
+use strict;
 
-use vars qw($VERSION $auto %used);
-$VERSION = "0.05";
-$auto = "Auto";
-%used = ();
+# Switch selecting XS or pure perl
+use vars qw($VERSION @ISA @implementors);
+$VERSION = "0.08";
 
-use AutoLoader qw(AUTOLOAD);
-
-use constant DEBUG => 0;
-
-sub _use {
-    my $name = shift();
-    $name =~ s|::|/|g;
-    print STDERR "require Heap/Simple/$name.pm\n" if DEBUG;
-    return require "Heap/Simple/$name.pm";
-}
-
-my %order = ("<"  => "Number",
-             ">"  => "NumberReverse",
-             "lt" => "String",
-             "gt" => "StringReverse",
-             );
-sub _order {
-    my ($self, $order) = @_;
-    # Default order if nothing specified
-    $order = "<" unless defined($order) && $order ne "";
-    my $name;
-    if (ref($order) eq "CODE") {
-        $self->[0][3] = $order;
-        $name = "Less";
-    } else {
-        $name = $order{$order} || croak "Unsupported order '$order'";
+unless (@ISA) {
+    @implementors = qw(Heap::Simple::XS(0.01) Heap::Simple::Perl(0.07))
+        unless @implementors;
+    for my $i (@implementors) {
+        my $plugin = $i;
+        my $version = $plugin =~ s/\((.+)\)\z// ? $1 : undef;
+        my $p = "$plugin.pm";
+        $p =~ s!::!/!g;
+        if (eval { require $p; 1 }) {
+            *new = $plugin->can("new") ||
+                croak "Package '$plugin' does not support 'new'";
+            if (defined $version) {
+                if (defined(my $v = $plugin->VERSION)) {
+                    if ($v < $version) {
+                        carp "Need '$plugin' version $version, found version $v";
+                        next;
+                    }
+                } else {
+                    croak "Could not determine the version of '$plugin'";
+                }
+            }
+            @ISA = ($plugin);
+            last;
+        }
     }
-    $used{$name} ||= _use($name);
-    return $name;
-}
-
-sub _elements {
-    my ($self, $elements) = @_;
-    $elements = ["Key"] unless defined($elements);
-    $elements = [$elements] if ref($elements) eq "";
-    croak "option elements is not an array reference" unless
-        ref($elements) eq "ARRAY";
-    croak "option elements has no type defined at index 0" unless
-        defined($elements->[0]);
-    my $name = ucfirst(lc($elements->[0]));
-    $used{$name} ||= _use($name);
-    return (__PACKAGE__ . "::$name")->_elements($self, $name, $elements);
-}
-
-sub _init {
-}
-
-sub new {
-    my ($class, %options) = @_;
-    # note: the array starts at elements 1 to make the subscripting
-    # operations (much!) cleaner.
-    # So elements 0 is used for associated data
-    my $self = bless [[]], $class;
-    # We temporarily bless $self into $class so you can play OO games with it
-    my @order    = $self->_order($options{order});
-    my @elements = $self->_elements($options{elements});
-    my $gclass = join("::", $class, $auto, @order, @elements);
-    no strict "refs";
-    @{"${gclass}::ISA"} = (__PACKAGE__ . "::$elements[0]",
-                           __PACKAGE__ . "::$order[0]",
-                           $class) unless @{"${gclass}::ISA"};
-    print STDERR "Generated class $gclass\n" if DEBUG;
-    # Now rebless the result in its final generated class
-    bless $self, $gclass;
-    $self->[0][1] = exists($options{infinity}) ?
-        $options{infinity} : $self->_INF;
-    $self->[0][4] = $options{user_data} if defined($options{user_data});
-    $self->_init(\%options);
-    return $self;
-}
-
-sub _ELEMENTS_PREPARE {
-    return "";
-}
-
-sub _ORDER_PREPARE {
-    return "";
-}
-
-sub _PREPARE {
-    my $self = shift;
-    return join("", $self->_ORDER_PREPARE, $self->_ELEMENTS_PREPARE);
-}
-
-sub _VALUE {
-    return $_[1];
-}
-
-sub _WRAPPER {
-    return $_[2];
-}
-
-sub _INF {
-    return;
-}
-
-sub _MAKE_KEY {
-    my ($self, $key, $value) = @_;
-    return "$key " . $self->_KEY($value);
-}
-
-my $balanced;
-# String with balanced parenthesis
-$balanced = qr{[^()]*(?:\((??{$balanced})\)[^()]*)*};
-sub _make {
-    my $self  = shift;
-
-    die "Cannot determine caller class from '$self'" unless ref($self);
-    my $subroutine = (caller(1))[3];
-    $subroutine =~ /(::[^:]*)\z/ || die "Cannot parse caller '$subroutine'";
-    $subroutine = ref($self) . $1;
-    print STDERR "Sub: $subroutine\n" if DEBUG;
-
-    my $string = shift;
-    # Very simple macro expander
-    1 while $string =~ s/(_\w+)\(($balanced)\)/$self->$1(split ',', $2)/eg;
-    print STDERR "Code: $string\n" if DEBUG;
-    no warnings 'redefine';
-    no strict 'refs';
-    *$subroutine = eval $string;
-    die $@ if $@;
-}
-
-sub count {
-    return $#{+shift};
-}
-
-sub clear {
-    $#{+shift} = 0;
+    @ISA || croak "Can't locate ", join(" nor ", @implementors), " in \@INC. This probably means you didn't install any of them. (\@INC contains: @INC)";
 }
 
 1;
-
 __END__
 
-sub insert {
-    my $self = shift;
-    if ($self->_KEY("") eq "") {
-        $self->_make('sub {
-    my ($self, $key) = @_;
-    _ORDER_PREPARE()
-    my $i = @$self;
-    $i = $i >> 1 while $i > 1 && _SMALLER($key, ($self->[$i] = $self->[$i >> 1]));
-    $self->[$i] = $key}');
-    } else {
-        $self->_make('sub {
-    my ($self, $value) = @_;
-    _PREPARE()
-    _MAKE_KEY(my $key =, $value);
-    my $i = @$self;
-    $i = $i >> 1 while
-        $i > 1 && _SMALLER($key, _KEY(($self->[$i] = $self->[$i >> 1])));
-    $self->[$i] = _WRAPPER($key, $value)}');
-    }
-    $self->insert(@_);
-}
-
-sub extract_upto {
-    my $self = shift;
-    $self->_make('sub {
-    my ($self, $border) = @_;
-    _PREPARE()
-    my @result;
-    push(@result, $self->extract_min) until
-        @$self <= 1 || _SMALLER($border, _KEY($self->[1]));
-    return @result
-}');
-    $self->extract_upto(@_);
-}
-
-sub extract_min {
-    my $self = shift;
-    $self->_make('sub {
-    my $self = shift;
-    if (@$self <= 2) {
-        return _VALUE(pop(@$self)) if @$self == 2;
-        croak "heap underflow";
-    }
-    my $min = _VALUE($self->[1]);
-    _PREPARE()
-    my $key = _KEY($self->[-1]);
-    my $n = @$self-2;
-    my $i = 1;
-    my $l = 2;
-    while ($l < $n) {
-        if (_SMALLER(_KEY($self->[$l]), $key)) {
-            if (_SMALLER(_KEY($self->[$l+1]), _KEY($self->[$l]))) {
-                $self->[$i] = $self->[$l+1];
-                $i = $l+1;
-            } else {
-                $self->[$i] = $self->[$l];
-                $i = $l;
-            }
-        } elsif (_SMALLER(_KEY($self->[$l+1]), $key)) {
-            $self->[$i] = $self->[$l+1];
-            $i = $l+1;
-        } else {
-            last;
-        }
-        $l = $i * 2;
-    }
-    if ($l == $n && _SMALLER(_KEY($self->[$l]), $key)) {
-        $self->[$i] = $self->[$l];
-        $i = $l;
-    }
-    $self->[$i] = pop(@$self);
-    return $min
-}');
-    $self->extract_min(@_);
-}
-
-sub min_key {
-    my $self = shift;
-    if ($self->can("_QUICK_KEY")) {
-        $self->_make('sub {
-    my $self = shift;
-    return @$self > 1 ? _QUICK_KEY($self->[1]) :
-        defined($self->[0][1]) ? $self->[0][1] : croak "Heap empty"
-}');
-    } else {
-        $self->_make('sub {
-    my $self = shift;
-    return defined($self->[0][1]) ? $self->[0][1] : croak "Heap empty" if
-        @$self <= 1;
-    _ELEMENTS_PREPARE()
-    return _KEY($self->[1])
-}');
-    }
-    $self->min_key(@_);
-}
-
-sub first_key {
-    my $self = shift;
-    if ($self->can("_QUICK_KEY")) {
-    $self->_make('sub {
-        my $self = shift;
-    return if @$self <= 1;	# avoid autovivify
-    return _QUICK_KEY($self->[1])
-}');
-    } else {
-    $self->_make('sub {
-        my $self = shift;
-    return if @$self <= 1;	# avoid autovivify
-    _ELEMENTS_PREPARE()
-    return _KEY($self->[1])
-}');
-    }
-    return $self->first_key(@_);
-}
-
-sub first {
-    my $self = shift;
-    if ($self->_VALUE("") eq "") {
-        $self->_make('sub {
-    return shift->[1]
-}');
-    } else {
-        $self->_make('sub {
-    my $self = shift;
-    return if @$self <= 1;	# avoid autovivify
-    return _VALUE($self->[1])
-}');
-    }
-    return $self->first(@_);
-}
-
-# This code is pretty weak. Often best to override for the mement
-sub key {
-    my $self = shift;
-    if ($self->_KEY("") eq "") {
-        $self->_make('sub {
-    return $_[1]}');
-    } else {
-        $self->_make('sub {
-    my $self = shift;
-    return _QUICK_KEY(shift)
-}');
-    }
-    return $self->key(@_);
-}
-
-sub keys {
-    my $self = shift;
-    if($self->_KEY("") eq "") {
-        $self->_make('sub {
-    my $self = shift;
-    return @$self[1..$#$self]}');
-    } else {
-        $self->_make('sub {
-    my $self = shift;
-    _ELEMENTS_PREPARE()
-    return map _KEY($_), @$self[1..$#$self]}');
-    }
-    return $self->keys(@_);
-}
-
-sub values {
-    my $self = shift;
-    if($self->_VALUE("") eq "") {
-        $self->_make('sub {
-    my $self = shift;
-    return @$self[1..$#$self]}');
-    } else {
-        $self->_make('sub {
-    my $self = shift;
-    return map _VALUE($_), @$self[1..$#$self]}');
-    }
-    return $self->values(@_);
-}
-
-sub user_data {
-    if (@_ > 1) {
-        my $self = shift;
-        my $old = $self->[0][4];
-        $self->[0][4] = shift;
-        return $old;
-    }
-    return shift->[0][4];
-}
-
-sub infinity {
-    if (@_ > 1) {
-        my $self = shift;
-        my $old = $self->[0][1];
-        $self->[0][1] = shift;
-        return $old;
-    }
-    return shift->[0][1];
-}
-
 =head1 NAME
 
 Heap::Simple - Fast and easy to use classic heaps
@@ -354,18 +54,21 @@ Heap::Simple - Fast and easy to use classic heaps
     # Put data in a "Object" or "Any" heap with a given key
     $heap->key_insert($key, $element);
 
-    # Extract data
-    $element = $heap->extract_min;
+    # Extract the top value
+    $element = $heap->extract_top;	# croaks on an empty heap
+    $element = $heap->extract_first;	# returns undef on an empty heap
+
+    # Get the top value but leave it in the heap
+    $element = $heap->top;		# croaks on an empty heap
+    $element = $heap->first;		# returns undef on an empty heap
 
     # Extract all data whose key is not above a given value
     @elements = $heap->extract_upto($max_key);
 
-    # Look which data is first without extracting
-    $element = $heap->first;
-
-    # Find the lowest value in the heap
-    $min_key = $heap->first_key;  # returns undef   on an empty heap
-    $min_key = $heap->min_key;	  # return infinity on an empty heap
+    # Find the top key in the heap
+    $top_key = $heap->top_key;	  	# return infinity on an empty heap
+					# croaks if there's no infinity
+    $top_key = $heap->first_key;  	# returns undef   on an empty heap
 
     # Find the number of elements
     $count = $heap->count;
@@ -395,6 +98,68 @@ Heap::Simple - Fast and easy to use classic heaps
     $key_method   = $heap->key_method;
     $key_function = $heap->key_function;
 
+    # Return the value of several things that were set in new:
+    $wrapped      = $heap->wrapped;
+    $max_count    = $heap->max_count;
+    $can_die      = $heap->can_die;
+    $dirty        = $heap->dirty;
+    $order	  = $heap->order;
+    @elements     = $heap->elements;
+    $elements     = $heap->elements;
+
+    # Move all elements from $heap2 to $heap1
+    $heap1->absorb($heap2);	# As if doing a repeated $heap1->insert
+    $heap1->key_absorb($heap2);	# As if doing a repeated $heap1->key_insert
+
+    # Which class does the actual work ?
+    $implementation = Heap::Simple->implementation;
+
+=head1 EXAMPLE1
+
+When key and value are kept separate:
+
+    use Heap::Simple;
+    my $heap = Heap::Simple->new(elements => "Any");
+
+    $heap->key_insert(8, "bar");
+    $heap->key_insert(5, "foo");
+
+    # This will print foo (5 is the top key)
+    print "First value is ", $heap->extract_top, "\n";
+
+    $heap->key_insert(7, "baz");
+
+    # This will print baz (7 is the top key)
+    print "Next value is ", $heap->extract_top, "\n";
+    # This will print bar (8 is now the top key)
+    print "Next value is ", $heap->extract_top, "\n";
+
+=head1 EXAMPLE2
+
+When the key is part of the value:
+
+    # This is purely for display, ignore it
+    use Data::Dumper;
+    $Data::Dumper::Indent = 0;
+    $Data::Dumper::Terse = 1;
+
+    # Real code starts here
+    use Heap::Simple;
+    my $heap = Heap::Simple->new(elements => "Array");
+
+    $heap->insert([8, "bar"]);
+    $heap->insert([5, "foo"]);
+
+    # This will print [5, foo] (5 is the top key)
+    print "First value is ", Dumper($heap->extract_top), "\n";
+
+    $heap->insert([7, "baz"]);
+
+    # This will print [7, baz] (7 is the top key)
+    print "Next value is ", Dumper($heap->extract_top), "\n";
+    # This will print [8, bar] (8 is now the top key)
+    print "Next value is ", Dumper($heap->extract_top), "\n";
+
 =head1 DESCRIPTION
 
 A heap is a partially sorted structure where it's always easy to extract the
@@ -416,13 +181,26 @@ object, that's what you will get back on extract. It's also ok to keep
 references to the elements around and make changes to them while they are
 in the heap as long as you don't change the key.
 
+Actually, Heap::Simple itself is just a loader for the code that will
+actually implement the functionality mentioned above. You will need to
+install something like L<Heap::Simple::XS|Heap::Simple::XS> or
+L<Heap::Simple::Perl|Heap::Simple::Perl> to be able to actually do anything.
+
 =head1 EXPORT
 
 None.
 
 =head1 METHODS
 
-=over 4
+New has a lot of ways to specify element types, but the right choices
+follows quite directly from the data you'll put in the heap. If the key
+is part of the data (or easily derived from the data), choose an element
+type that tells how to get the key out of the data, and insert elements
+using L<insert|/"insert">. If the key is independent from the data or
+you want to avoid repeated key calculations, use the L<Any|/"Any"> element
+type and insert elements using L<key_insert|/"key_insert">.
+
+=over
 
 =item X<new>my $heap = Heap::Simple->new
 
@@ -434,7 +212,7 @@ You could for example use this to print a list of numbers from low to high:
 
     my $heap = Heap::Simple->new;
     $heap->insert($_) for 8, 3, 14, -1, 3;
-    print $heap->extract_min, " " for 1..$heap->count;
+    print $heap->extract_top, " " for 1..$heap->count;
     print "\n";
     # Will print: -1 3 3 8 14
 
@@ -452,7 +230,7 @@ Several options can help you with that:
 
 =over 2
 
-=item X<order>order => $order
+=item X<new_order>order => $order
 
 $order indicates what operation is used to compare keys. Supported orders are:
 
@@ -473,10 +251,7 @@ The default infinity for this order is +inf.
 =item E<gt>
 
 Indicates that keys are compared as numbers, and extraction is highest value
-first. This means that methods like L<extract_min|"extract_min"> become
-rather confusing in name, since they extract the maximum in the sense of
-the numeric value (but it's still the smallest value in terms of the
-abstract order relation).
+first.
 
 Repeating the example with this order gives:
 
@@ -484,7 +259,7 @@ Repeating the example with this order gives:
 
     my $heap = Heap::Simple->new(order => ">");
     $heap->insert($_) for 8, 3, 14, -1, 3;
-    print $heap->extract_min, " " for 1..$heap->count;
+    print $heap->extract_top, " " for 1..$heap->count;
     print "\n";
     # Will print: 14 8 3 3 -1
 
@@ -499,7 +274,7 @@ value first. So we could modify the "<" example to:
 
     my $heap = Heap::Simple->new(order => "lt");
     $heap->insert($_) for "ate", 8, 3, "zzzz", 14, -1, 3, "at";
-    print $heap->extract_min, " " for 1..$heap->count;
+    print $heap->extract_top, " " for 1..$heap->count;
     print "\n";
     # Will print: -1 14 3 3 8 at ate zzzz
 
@@ -517,16 +292,16 @@ The standard example now becomes:
 
     my $heap = Heap::Simple->new(order => "gt");
     $heap->insert($_) for "ate", 8, 3, "zzzz", 14, -1, 3, "at";
-    print $heap->extract_min, " " for 1..$heap->count;
+    print $heap->extract_top, " " for 1..$heap->count;
     print "\n";
     # Will print: zzzz ate at 8 3 3 14 -1
 
 The default infinity for this order is "" (the empty string)
 
-=item $code_reference
+=item X<CODE>$code_reference
 
-If your keys are completely weird things, ordered neither as numbers nor as 
-strings and you need a special compare function, you can use this general 
+If your keys are completely weird things, ordered neither as numbers nor as
+strings and you need a special compare function, you can use this general
 ordering type.
 
 Every time two keys need to be compared, the given code reference will be
@@ -550,20 +325,20 @@ Example:
 
     my $heap = Heap::Simple->new(order => \&more);
     $heap->insert($_) for 8, 3, 14, -1, 3;
-    print $heap->extract_min, " " for 1..$heap->count;
+    print $heap->extract_top, " " for 1..$heap->count;
     print "\n";
     # Will print: 14 8 3 3 -1
 
 The code reference will be called many times during normal heap operations
 (O(log n) times for a single insert or extract on a size n heap), so only use
-this order type within reason. Usually it's better to precalculate some number 
+this order type within reason. Usually it's better to precalculate some number
 or string representation of some sort of key and use normal compares on these.
-You can use the L<Any element type|"Any"> and L<key_insert|"key_insert"> to 
+You can use the L<Any element type|"Any"> and L<key_insert|"key_insert"> to
 wrap the precalculated key with the corresponding element, or you can delegate
 the key calculation to the L<insert|"insert"> method and use one of the
 L<Method|"Method">, L<Object|"Object"> or L<Function|"Function"> element types.
 
-Here's an example of such "fake" keys: 
+Here's an example of such "fake" keys:
 
     # "human" sorting mixed strings
     use Heap::Simple;
@@ -573,12 +348,12 @@ Here's an example of such "fake" keys:
         $str =~ s|(0*)(\d+)|pack("AN/A*N", "0", $2, length($1))|eg;
         return $str;
     }
-       
+
     my $heap = Heap::Simple->new(order => "lt",
                                  elements => [Function => \&key]);
     $heap->insert($_) for qw(Athens5.gr Athens40.gr
                              Amsterdam51.nl Amsterdam5.nl amsterdam20.nl);
-    print $heap->extract_min, "\n" for 1..$heap->count;
+    print $heap->extract_top, "\n" for 1..$heap->count;
     # This will print:
     Amsterdam5.nl
     amsterdam20.nl
@@ -588,7 +363,7 @@ Here's an example of such "fake" keys:
 
 =back
 
-=item X<elements>elements => $element_type
+=item X<new_elements>elements => $element_type
 
 This option describes what sort of elements we will store in the heap.
 The only reason the module needs to know this is to determine how to access
@@ -607,18 +382,21 @@ The following element types are currently supported:
 
 =over 2
 
-=item X<Key>["Key"]
+=item X<Scalar>["Scalar"]
 
 Indicates that the elements are the keys themselves. This is the default if no
 elements option is provided. So the constructor in the previous example could
 also have been written as:
 
     my $heap = Heap::Simple->new(order => "lt",
-                                 elements => ["Key"]);
+                                 elements => ["Scalar"]);
 
 or in the simplified notation:
 
-    my $heap = Heap::Simple->new(order => "lt", elements => "Key");
+    my $heap = Heap::Simple->new(order => "lt", elements => "Scalar");
+
+This element type used to be called C<Key>, and that name is still accepted
+for backward compatibility.
 
 =item X<Array>[Array => $index]
 
@@ -634,14 +412,14 @@ use this to for example print the values of a hash ordered by key:
         $heap->insert([$key, $val]);
     }
     for (1..$heap->count) {
-        print $heap->extract_min->[1], "\n";
+        print $heap->extract_top->[1], "\n";
     }
 
 You can always use something like [$key, @data] to pair up keys and data,
 so the "Array" element type is rather generally useful (but see the
 L<Object|"Object"> and L<Any|"Any"> element types for another way to pair
-keys with data). Since it's so common to have the key in the first position, 
-you may in fact drop the index in that case, so the constructor in the 
+keys with data). Since it's so common to have the key in the first position,
+you may in fact drop the index in that case, so the constructor in the
 previous example could also be written as:
 
     my $heap = Heap::Simple->new(order => "lt",
@@ -676,10 +454,10 @@ Redoing the Array example in Hash style gives:
         $heap->insert({tag => $key, value => $val});
     }
     for (1..$heap->count) {
-        print $heap->extract_min->{value}, "\n";
+        print $heap->extract_top->{value}, "\n";
     }
 
-In case the elements you want to store are hashes (or hash based objects and 
+In case the elements you want to store are hashes (or hash based objects and
 you are prepared to break the object encapsulation), this element type is also
 very nice. If for example the value on which you want to order is a number
 with key "price", you could use:
@@ -736,7 +514,7 @@ calculate the key, but do it yourself before the insert, and then use
 L<key_insert|"key_insert">. In fact, if you never use plain L<insert|"insert">
 at all, you don't even have to bother passing a method name (though in that
 case the fact that the thing you store is an object is pretty irrelevant and
-it's probable more natural to use the L<Any element type|"Any">.
+it's probable more natural to use the L<Any element type|"Any">).
 
 =item X<Function>[Function => $code_reference]
 
@@ -799,12 +577,71 @@ Or we can use it to simplify the hash sort on key example a bit:
         $heap->key_insert($key, $val);
     }
     for (1..$heap->count) {
-        print $heap->extract_min, "\n";
+        print $heap->extract_top, "\n";
     }
 
 =back
 
-=item user_data => $user_data
+=item X<new_max_count>max_count => $natural_number
+
+Normally a heap can contain any number of elements. By passing a positive
+integer as argument to max_count, you tell the heap that it should never
+contain more values than that. If you try to insert something new when the
+maximum is reached, the lowest element (with respect to the current order) is
+dropped (the thing just being inserted is among the candidates for dropping).
+
+A max count of 0 may or may not be supported depending on the implementor.
+
+You can for example use this to determine the highest three values in an array:
+
+    use Heap::Simple;
+
+    my @array = qw(19 3 7 -5 3 18 1);
+
+    my $heap = Heap::Simple->new(max_count => 3);
+    $heap->insert($_) for @array;
+    print "The three highest values are: ", join(", " => $heap->values), "\n";
+
+    # Will print: The three highest values are: 7, 19, 18
+
+=item X<new_can_die>can_die => $bool
+
+If you use magic values, overload, L<order functions|"CODE"> or
+L<key access functions|"new_elements">, then it's possible for external perl
+code to run when you do heap operations like L<insert|"insert"> or
+L<extract_top|"extract_top">. If these throw an exception, the heap can
+be left in an incorrect half changed state.
+
+If you give a true value to can_die, the code for these operations will be
+changed so that they will properly recover by undoing what just got changed
+(so a failing operation becomes a no-op). This however will slow down these
+operations somewhat, so the default is actually false (most of the time getting
+exceptions during the heap operations is impossible anyways).
+
+This is a per heap option, so only those heaps that actually set this will
+see a slowdown.
+
+All operations that don't change the heap (like L<count|"count"> or 
+L<top|"top">) are always safe.
+
+Note that all change operations always assume you won't recursively cause
+another change while they are running. If you do that, all bets for consistent
+heaps are are off, even if you set this option.
+
+=item X<dirty>dirty => $bool
+
+Giving this option a true value means that the code may make optimistic
+assumptions to gain more speed. These can be things like for example ignoring
+overloads, casting all numbers to doubles, ignoring locale for string order,
+caching keys etc. The particular optimizations done under dirty will be safe
+for most applications though. See the documentation for a particular
+implementor (like L<Heap::Simple::XS|Heap::Simple::XS> or
+L<Heap::Simple::Perl|Heap::Simple::Perl>) for what the dirty option does for
+that package.
+
+The default is no dirty optimizations.
+
+=item X<new_user_data>user_data => $user_data
 
 You can associate one scalar worth of user data with any heap. This option
 allows you to set its value already at object creation. You can use the
@@ -816,16 +653,16 @@ If this option is not given, the heap starts with "undef" associated to it.
     print $heap->user_data, "\n";
     # prints foo
 
-=item infinity => $infinity
+=item X<new_infinity>infinity => $infinity
 
 Associates $infinity as the highest possible key with the created heap.
 ($infinity may or may not be a possible key itself).
 Setting it to "undef" means there is no infinity associated with the heap.
 
-The default value depends on the L<order|"order"> relation that was
+The default value depends on the L<order|"new_order"> relation that was
 specified.
 
-Usually you can just forget about this option. Only L<min_key|"min_key">
+Usually you can just forget about this option. Only L<top_key|"top_key">
 really cares.
 
 =back
@@ -842,24 +679,47 @@ $element as you inserted, including a possible L<blessing|perlfunc/"bless">.
 =item X<key_insert>$heap->key_insert($key, $element)
 
 Inserts the $element in the heap ordered by the given $key. Since in this case
-the key must be stored seperately from the element, this only works for
+the key must be stored seperately from the element, this method only exists for
 L<"Object"|"Object"> and L<"Any"|"Any"> heaps.
 
 On extraction you get back exactly the same $element as you inserted,
 including a possible L<blessing|perlfunc/"bless">.
 
-=item X<extract_min>$element = $heap->extract_min
+=item X<extract_top>$element = $heap->extract_top
 
-For all elements in the heap, find the one with the lowest key, remove it from
-the heap and return it.
+For all elements in the heap, find the top one (the one that is "lowest" in the
+order relation), remove it from the heap and return it.
+
+This method used to be called C<"extract_min"> instead of C<"extract_top">.
+The old name is still supported but is deprecated.
 
 Throws an exception if the heap is empty.
 
-=item X<extract_upto>@elements = $heap->extract_upto($value)
+=item X<extract_first>$element = $heap->extract_top
+
+Like L<extract_top|"extract_top">, but if the heap is empty it will
+return undef (in scalar context).
+
+=item X<top>$element = $heap->top
+
+For all elements in the heap, find the top one (the one that is "lowest" in the
+order relation) and return it (without removing it from the heap)..
+
+Throws an exception if the heap is empty.
+
+=item X<first>$element = $heap->first
+
+For all elements in the heap, find the top one (the one that is "lowest" in the
+order relation) and return it (without removing it from the heap)..
+
+Returns undef if the heap is empty.
+
+=item X<extract_upto>@elements = $heap->extract_upto($max_key)
 
 Finds all elements in the heap whose key is not above $value and removes them
-from the heap. The list of removed element is returned ordered by key value
-(low to high).
+from the heap (so elements with key equal to $max_key get extracted too).
+The list of removed elements is returned ordered by key value (low to high
+with repect to the heap order).
 
 Returns an empty list for the empty heap.
 
@@ -872,13 +732,18 @@ Example:
     print join(", ", $heap->extract_upto(3)), "\n";
     # prints -1, 3, 3
 
+This method will lose values in case of an exception even if
+L<can_die|"new_can_die"> is true (remember that exceptions of this type are
+only possible if you have a self coded key fetch or compare that can die, so
+this is normally irrelevant).
+
 =item X<min>$element = $heap->first
 
 For all elements in the heap, find the one with the lowest key and return it.
-Returns undef in case the heap is empty. The contents of the heap remain
-unchanged.
+Returns undef (in scalar context) in case the heap is empty. The contents of
+the heap remain unchanged.
 
-Since the data returned from a non-empty heap can usually not be undef, you
+Since the data returned from a non-empty heap can often not be undef, you
 could use this method to check if a heap is empty, but it's probably more
 natural to use L<count|"count"> for that.
 
@@ -891,10 +756,10 @@ Example:
     print $heap->first, "\n";
     # prints -1
 
-=item X<first_key>$min_key = $heap->first_key
+=item X<first_key>$top_key = $heap->first_key
 
 Looks for the lowest key in the heap and returns its value. Returns undef
-in case the heap is empty
+(in scalar context) in case the heap is empty
 
 Example:
 
@@ -905,7 +770,7 @@ Example:
     print $heap->first_key, "\n";
     # prints -1
 
-=item X<min_key>$min_key = $heap->min_key
+=item X<top_key>$top_key = $heap->top_key
 
 Looks for the lowest key in the heap and returns its value. Returns the highest
 possible value (the infinity for the chosen order) in case the heap is empty.
@@ -917,18 +782,26 @@ Example:
 
     my $heap = Heap::Simple->new;
     $heap->insert($_) for 8, 3, 14, -1, 3;
-    print $heap->min_key, "\n";
+    print $heap->top_key, "\n";
     # prints -1
 
+This method used to be called "min_key" instead of "top_key". The old name is
+still supported but is deprecated.
+
 =item X<key>$key = $heap->key($value)
 
 Calculates the key corresponding to $value in the same way as the internals
 of $heap would. Can fail for L<Object|"Object"> and L<Any|"Any"> element
 types if there was no method or function given on heap creation.
 
+Notice that this does not access the elements in the heap in any way.
+In particular, it's B<not> looking for $value in the heap hoping to match its
+key.
+
 =item X<count>$count = $heap->count
 
 Returns the number of elements in the heap.
+This is an constant time operation, it doesn't really need to count anything.
 
     use Heap::Simple;
 
@@ -943,16 +816,27 @@ Removes all elements from the heap.
 
 =item X<keys>@keys = $heap->keys
 
-Returns the keys of all elements in the heap in some unspecified order
-(so don't expect them to be ordered). This may imply a lot of function
-calls if getting the key from an element implies a function call
-(as it does for the L<Method|"Method"> and L<Function|"Function"> element
-types, but not for the L<Object|"Object"> and L<Any|"Any"> element types).
+Returns the keys of all elements in the heap in some heap order.
+This means that the element at index n is not bigger (in the heap order)
+than the element at index 2*n+1 and 2*n+2. So the top key will in fact be
+in the first position, but don't expect the whole list to be ordered.
+
+This method may imply a lot of function calls if getting the key from an
+element implies a function call (as it does for the L<Method|"Method"> and
+L<Function|"Function"> element types, but not for the L<Object|"Object"> and
+L<Any|"Any"> element types).
+
+Multiple calls to an unchanged heap will return the keys in the same order,
+which is also consistent with the order of L<values|"values">
 
 =item X<values>@values = $heap->values
 
-Returns all elements in the heap in some unspecified order
-(does not remove them from the heap).
+Returns all elements in the heap in some heap order with respect to the
+corresponding keys (see L<keys|"keys">).
+Does not remove the values from the heap.
+
+Multiple calls to an unchanged heap will return the values in the same order,
+which is also consistent with the order of L<keys|"keys">
 
 =item X<user_data>$user_data = $heap->user_data
 
@@ -967,42 +851,121 @@ Associates new user_data with the heap. Returns the old value.
 Queries the infinity value associated with the heap. Returns undef if there
 is none. The default infinity is implied by the chosen order relation.
 
-=item $old_infinity = $heap->user_data($new_infinity)
+=item $old_infinity = $heap->infinity($new_infinity)
 
 Associates a new infinity with the heap. Returns the old value.
 
-=item $key_index = $heap->key_index
+=item X<key_index>$key_index = $heap->key_index
 
 Returns the index of the key for L<array reference based heaps|"Array">.
 Doesn't exist for the other heap types.
 
-=item $key_name = $heap->key_name
+=item X<key_name>$key_name = $heap->key_name
 
 Returns the name of the key key for L<hash reference based heaps|"Hash">.
 Doesn't exist for the other heap types.
 
-=item $key_name = $heap->key_method
+=item X<key_method>$key_name = $heap->key_method
 
 Returns the name of the method to fetch the key from an object. Only exists
 for L<Method|"Method"> and L<Object|"Object"> based heaps.
 
-=item $key_function = $heap->key_function
+=item X<key_function>$key_function = $heap->key_function
 
 Returns the code reference of the function to fetch the key from an element.
 Only exists for L<"Function"|"Function"> and L<"Any"|"Any"> heaps.
 
+=item X<wrapped>$wrapped => $heap->wrapped
+
+Returns true if key and value are stored seperately (wrapped together in
+some internal container), nothing otherwise. This is the sufficient and
+necessary condition for L<key_insert|"key_insert"> to work, and will normally
+only be true for the L<"Any"|"Any"> and L<Object|"Object"> type heaps.
+
+=item X<max_count>$max_count => $heap->max_count
+
+Returns the maximum size of the heap, or infinity if there is no maximum (the
+default unless you used the L<max_count|"new_max_count"> option.
+
+=item X<can_die>$can_die = $heap->can_die
+
+Returns the L<can_die|"new_can_die"> setting for this heap.
+
+=item X<dirty>$dirty = $heap->dirty
+
+Returns the L<dirty|"new_dirty"> setting for this heap.
+
+=item X<order>$order = $heap->order
+
+Returns the L<order|"new_order"> setting for this heap (either C<E<lt>>, C<E<gt>>, C<lt>, C<gt> or a code reference).
+
+=item X<elements>@elements = $heap->elements
+
+Returns the L<elements|"new_elements"> setting for this heap. The first entry
+in the returned list is a string representing the type in canonical form
+(L<"Scalar"|"Scalar"> or L<"Array"|"Array"> etc) followed by any arguments
+that type needed (e.g. the key name for a L<Hash|"Hash"> type).
+
+=item X<elements>$elements = $heap->elements
+
+Like the list context version, but only returns the first entry (the canonical
+type name).
+
+=item X<absorb>$heap1->absorb($heap2)
+
+Take all elements from $heap2 and inserts them in $heap1, leaving $heap2 empty.
+Behaves a bit like:
+
+    $heap1->insert($_) for reverse $heap2->values;
+    $heap2->clear;
+
+except that it's much more efficient.
+
+If an exception is possible and gets raised during insert, the heaps will be
+left in a consistent state with a partial transfer completed on the condition
+that L<can_die|"new_can_die"> is set for $heap1 (the setting for $heap2 is
+irrelevant, the accesses there will always be done in a safe way)
+
+=item X<key_absorb>$heap1->key_absorb($heap2)
+
+Take all elements from $heap2 and key_inserts them in $heap1, leaving $heap2
+empty. Behaves a bit like:
+
+    my @values = $heap2->values;
+    my @keys   = $heap2->keys;
+    $heap1->key_insert(pop @keys, pop @values) while @values;
+    $heap2->clear;
+
+except that it's much more efficient. This is mainly meant for transfer between
+wrapped heap types (L<Any|"Any"> and L<Object|"Object"> (it avoids key
+recalculation). $heap1 must of course be a wrapped heap type.
+
+If an exception is possible and gets raised during insert, the heaps will be
+left in a consistent state with a partial transfer completed on the condition
+that L<can_die|"new_can_die"> is set for $heap1 (the setting for $heap2 is
+irrelevant, the accesses there will always be done in a safe way)
+
+=item X<implementation>$implementation = Heap::Simple->implementation
+
+Returns the package that does the actual work. That will probably be
+C<"Heap::Simple::XS"> or C<"Heap::Simple::Perl">.
+
 =back
 
 =head1 SEE ALSO
 
+L<Heap::Simple::Perl>,
+L<Heap::Simple::XS>
+
+Some other heap or heap-like classes that exist:
+
 L<Heap>,
-L<Heap::Priority>
+L<Heap::Priority>,
+L<Array::Heap2>
 
 =head1 AUTHOR
 
-Ton Hospel, E<lt>Heap::Simple@home.lunixE<gt>
-
-Parts are based on code by Joseph N. Hall (http://www.perlfaq.com/faqs/id/196)
+Ton Hospel, E<lt>Heap::Simple@ton.iguana.beE<gt>
 
 =head1 COPYRIGHT AND LICENSE
 
